from __future__ import annotations

import asyncio
from typing import Any, Callable, Sequence

from .asyncapi.models import PublishInfo
from .asyncapi.registry import PUBLISH_REGISTRY
from .broker import Broker
from .consumer import Consumer, ConsumerGroup, ForwardResponse
from .logger import LoggerMixin
from .models import CloudEvent
from .settings import ServiceSettings
from .types import TagMeta
from .utils import generate_instance_id


class Service(LoggerMixin):
    """Logical group of consumers. Provides group (queue) name and handles versioning"""

    def __init__(
        self,
        name: str,
        broker: Broker,
        title: str | None = None,
        version: str = "0.1.0",
        description: str = "",
        tags_metadata: list[TagMeta] | None = None,
        instance_id_generator: Callable[[], str] | None = None,
        base_event_class: type[CloudEvent] = CloudEvent,
        publish_info: Sequence[PublishInfo] = (),
        **context: Any,
    ):
        self.broker = broker
        self.name = name
        self.title = title or name.title()
        self.version = version
        self.description = description
        self.tags_metadata = tags_metadata or []
        self.id = (instance_id_generator or generate_instance_id)()
        self.consumer_group = ConsumerGroup()
        self.context = context
        self.base_event_class = base_event_class
        # TODO: task gathering?
        self._tasks: list[asyncio.Task] = []
        for p in publish_info:
            PUBLISH_REGISTRY[p.event_type.__name__] = p

    def subscribe(
        self,
        topic: str,
        *,
        name: str | None = None,
        timeout: int = 120,
        dynamic: bool = False,
        forward_response: ForwardResponse | None = None,
        **options,
    ):
        return self.consumer_group.subscribe(
            topic=topic,
            name=name,
            timeout=timeout,
            dynamic=dynamic,
            forward_response=forward_response,
            **options,
        )

    def add_consumer(self, consumer: Consumer):
        self.consumer_group.add_consumer(consumer)

    def add_consumer_group(self, consumer_group: ConsumerGroup) -> None:
        self.consumer_group.add_consumer_group(consumer_group)

    async def send(
        self,
        topic: str,
        type_: type[CloudEvent] | str = "CloudEvent",
        data: Any | None = None,
        **kwargs,
    ):
        if isinstance(type_, str):
            cls = self.base_event_class
            type_name = type_
        else:
            cls = type_
            type_name = type_.__name__

        message: CloudEvent = cls(
            content_type=self.broker.encoder.CONTENT_TYPE,
            type=type_name,
            topic=topic,
            data=data,
            source=self.name,
            **kwargs,
        )
        return await self.broker.publish(message)

    @property
    def consumers(self):
        return self.consumer_group.consumers

    async def publish(self, message: CloudEvent, **kwargs):
        if not message.source:
            message.set_source(self.name)
        return await self.broker.publish(message, **kwargs)

    async def start(self):
        await self.broker.connect()
        await self.broker.dispatch_before("service_start", self)
        for consumer in self.consumers.values():
            asyncio.create_task(self.broker.start_consumer(self, consumer))
        await self.broker.dispatch_after("service_start", self)

    async def stop(self, *args, **kwargs):
        await self.broker.dispatch_before("service_stop", self)
        await self.broker.disconnect()
        await self.broker.dispatch_after("service_stop", self)

    def run(self, *args, **kwargs):
        from .runner import ServiceRunner

        runner = ServiceRunner([self])
        runner.run(*args, **kwargs)

    @classmethod
    def from_settings(cls, settings: ServiceSettings, **kwargs: Any) -> Service:
        return cls(**settings.dict(), **kwargs)

    @classmethod
    def from_env(cls, **kwargs: Any) -> Service:
        return cls.from_settings(ServiceSettings(), **kwargs)



import os
import socket
import uuid


def str_uuid() -> str:
    return str(uuid.uuid4())


def generate_instance_id() -> str:
    pid = os.getpid()
    hostname = socket.gethostname()
    return f"{hostname}:{pid}"



from __future__ import annotations

from fastapi import FastAPI
from fastapi.responses import JSONResponse

from ..asyncapi.generator import get_async_api_spec
from ..asyncapi.models import AsyncAPI
from ..plugins import Service, ServicePlugin


class FastAPIServicePlugin(ServicePlugin):
    """Integration with fastapi, allows"""

    def __init__(
        self,
        service: Service,
        app: FastAPI,
        async_api_url: str | None = None,
        healthcheck_url: str | None = None,
    ):
        super().__init__(service)
        app.on_event("startup")(service.start)
        app.on_event("shutdown")(service.stop)

        if async_api_url:

            @app.get(async_api_url, response_model=AsyncAPI)
            def get_service_asyncapi():
                """Return service Async API specification"""
                return get_async_api_spec(service)

        if healthcheck_url:

            @app.get(healthcheck_url, response_class=JSONResponse)
            def get_broker_connection_status():
                try:
                    res = self.service.broker.is_connected
                    if not res:
                        raise ValueError
                    return JSONResponse({"status": "ok"})
                except Exception:
                    return JSONResponse(
                        {"detail": "Service Unavailable", "status": 503},
                        status_code=503,
                    )



from typing import List, Optional

from pydantic import BaseSettings, Field

from eventiq.utils.imports import ImportedType

from .middleware import Middleware


class BrokerSettings(BaseSettings):
    description: Optional[str] = None
    middlewares: Optional[List[Middleware]] = None
    encoder: Optional[ImportedType] = Field(None, env="BROKER_ENCODER_CLASS")


class ServiceSettings(BaseSettings):
    name: str
    title: Optional[str] = None
    version: str = "0.1.0"
    description: str = ""



from datetime import datetime, timedelta
from functools import partial
from typing import Any, Dict, Generic, Optional

from pydantic import AnyUrl, Extra, Field, validator
from pydantic.fields import PrivateAttr
from pydantic.generics import GenericModel

from .context import get_current_service
from .message import Message
from .types import ID, D
from .utils import str_uuid
from .utils.datetime import utc_now


class CloudEvent(GenericModel, Generic[D]):
    specversion: Optional[str] = "1.0"
    content_type: str = Field("application/json", alias="datacontenttype")
    id: ID = Field(default_factory=str_uuid)
    time: datetime = Field(default_factory=utc_now)
    topic: str = Field(..., alias="subject")
    type: Optional[str] = None
    source: Optional[str] = None
    data: D
    dataschema: Optional[AnyUrl] = None
    tracecontext: Dict[str, Any] = {}

    _raw: Optional[Any] = PrivateAttr()

    def __eq__(self, other):
        if not isinstance(other, CloudEvent):
            return False
        return self.id == other.id

    @validator("type", allow_reuse=True, always=True, pre=True)
    def get_type_from_cls_name(cls, v):
        return v or cls.__name__

    @property
    def raw(self) -> Message:
        if self._raw is None:
            raise AttributeError("raw property accessible only for incoming messages")
        return self._raw

    @property
    def service(self):
        return get_current_service()

    @property
    def log_context(self) -> Dict[str, Any]:
        return {"message_id": str(self.id)}

    def _set(self, name: str, value: Any):
        object.__setattr__(self, name, value)

    def set_source(self, value: str) -> None:
        self._set("source", value)

    def set_raw(self, value: Message):
        self._set("_raw", value)

    def dict(self, **kwargs: Any) -> Dict[str, Any]:
        kwargs.setdefault("by_alias", True)
        return super().dict(**kwargs)

    @classmethod
    def new(cls, obj: D, **kwargs: Any):
        return cls(data=obj, **kwargs)

    def copy(self, **kwargs):
        kwargs.setdefault(
            "exclude",
            {
                "id",
                "time",
            },
        )
        kwargs.setdefault("deep", True)
        return super().copy(**kwargs)

    @property
    def age(self) -> timedelta:
        return utc_now() - self.time

    def fail(self):
        self.raw.fail()

    class Config:
        use_enum_values = True
        allow_population_by_field_name = True
        extra = Extra.allow
        # events are immutable, however it's sometimes useful to set source or traceid for unpublished events
        allow_mutation = False


TopicField = partial(Field, const=True, alias="subject")



from ._version import __version__
from .asyncapi import publishes
from .broker import Broker
from .consumer import Consumer, ConsumerGroup, ForwardResponse, GenericConsumer
from .context import get_current_service
from .message import Message
from .middleware import Middleware
from .models import CloudEvent
from .plugins import BrokerPlugin, ServicePlugin
from .runner import ServiceRunner
from .service import Service
from .types import RawMessage

__all__ = [
    "__version__",
    "Broker",
    "BrokerPlugin",
    "Consumer",
    "ConsumerGroup",
    "CloudEvent",
    "ForwardResponse",
    "GenericConsumer",
    "Message",
    "Middleware",
    "RawMessage",
    "Service",
    "ServicePlugin",
    "ServiceRunner",
    "publishes",
    "get_current_service",
]



import asyncio
from typing import Sequence

from .service import Service


class ServiceRunner:
    """
    Service container for running multiple service instances in one process.
    :param services: Sequence of services to run
    """

    def __init__(self, services: Sequence[Service]) -> None:
        self.services = services

    def run(self, use_uvloop: bool = False, **kwargs) -> None:
        import aiorun

        aiorun.run(
            self._run(), shutdown_callback=self._stop, use_uvloop=use_uvloop, **kwargs
        )

    async def _run(self):
        await asyncio.gather(*[s.start() for s in self.services])

    async def _stop(self, *args, **kwargs):
        await asyncio.gather(*[s.stop() for s in self.services])



import inspect
import logging
from typing import Type, Union

from pythonjsonlogger import jsonlogger

from .context import get_current_message
from .utils.datetime import utc_now


class EventiqJsonFormatter(jsonlogger.JsonFormatter):
    def add_fields(self, log_record, record, message_dict):
        super().add_fields(log_record, record, message_dict)
        if "timestamp" not in log_record:
            log_record["timestamp"] = utc_now()

        msg = get_current_message()
        if msg:
            log_record.update(msg.log_context)

        log_record["level"] = record.levelname.upper()


def setup_logging(level, fmt: str = "%(name) %(level) %(message)") -> None:
    handler = logging.StreamHandler()
    handler.setFormatter(EventiqJsonFormatter(fmt))
    logging.basicConfig(handlers=[handler], level=level)


def get_logger(module: str, name: Union[str, Type]) -> logging.Logger:
    if inspect.isclass(name):
        name = name.__name__

    return logging.getLogger(f"{module}.{name}")


class LoggerMixin:
    logger: logging.Logger

    def __init_subclass__(cls, **kwargs):
        cls.logger = get_logger(__name__, cls)
